""" Handles the GPT API and the conversation state. """
import json
import time
from pathlib import Path
from typing import Optional, Union, List

from PIL.Image import Image

from justai.tools.cache import cached_llm_response
from justai.model.message import Message
from justai.models.modelfactory import ModelFactory

ImageInput = Optional[Union[
    List[str],
    List[bytes],
    List[Image],
    str,
    bytes,
    Image
]]

class Model:
    def __init__(self, model_name: str, **kwargs):
        
        # Model parameters
        self.model = ModelFactory.create(model_name, **kwargs)

        # Parameters to save the current conversation
        self.save_dir = Path(__file__).resolve().parent / 'saves'
        self.message_memory = 20  # Number of messages to remember. Limits token usage.
        self.messages = []  # List of Message objects
        self.tools = []  # List of tools to use / functions to call
        self.functions = {}  # The actual functions to call with key the name of the function and as value the function

        self.input_token_count = 0
        self.output_token_count = 0
        self.last_response_time = 0
        
        self.logger = None
        
    def __setattr__(self, name, value):
        if name not in self.__dict__ and hasattr(self, 'model') and name in self.model.model_params:
            # Not an existing property model but a model_params property. Set it in model_params
            self.model.model_params[name] = value
        else:
            # Update the property as intended
            super().__setattr__(name, value)

    @classmethod
    def from_json(cls, model_name, model_data, **kwargs):
        """ Creates an model from a json string. Usefull in stateless environments like a web page """
        model = cls(model_name, **kwargs)
        dictionary = json.loads(model_data)
        for key, value in dictionary.items():
            match key:
                case 'messages':
                    model.messages = [Message.from_dict(m) for m in value]
                case _:
                    model.__setattr__(key, value)
        return model

    def set_api_key(self, key: str):
        """ Used when using Aigent from a browser where the user has to specify a key """
        self.model.set('api_key', key)

    @property
    def system(self):  # This function can be overwritten by child classes to make the system message dynamic
        return self.model.system_message

    @system.setter
    def system(self, value):
        self.model.system_message = value

    @property
    def cached_prompt(self): 
        if hasattr(self.model, 'cached_prompt'):
            return self.model.cached_prompt
        raise AttributeError("Model does not support cached_prompt")

    @cached_prompt.setter
    def cached_prompt(self, value):
        if hasattr(self.model, 'cached_prompt'):
            self.model.cached_prompt = value
        else:
            raise AttributeError("Model does not support cached_prompt")

    @property
    def cache_creation_input_tokens(self):
        if hasattr(self.model, 'cache_creation_input_tokens'):
            return self.model.cache_creation_input_tokens
        raise AttributeError("Model does not support cache_creation_input_tokens")
    
    @property
    def cache_read_input_tokens(self):
        if hasattr(self.model, 'cache_read_input_tokens'):
            return self.model.cache_read_input_tokens
        raise AttributeError("Model does not support cache_read_input_tokens")
        
    def reset(self):
        self.messages = []

    def append_messages(self, prompt: str, images: ImageInput = None):
        if images:
            if not isinstance(images, list):
                images = [images]
        else:
            images = []
        self.messages.append(Message('user', prompt, images))
        return self.messages

    def add_function(self, func: callable, description: str, parameters: dict, required_parameters=None):
        """ Adds a function to the model.
        name: name of the function
        description: description of the function
        parameters: dictionary with parameter name as key and parameter type as value """
        # Basic mapping from Python types to JSON schema types
        type_mapping = {
            int: "integer",
            str: "string",
            float: "number",
            bool: "boolean",
            list: "array",
            dict: "object"
        }

        tool = {
            "type": "function",
            "function": {
                "name": func.__name__,
                "description": description,
                "parameters": {
                    "type": "object",
                    "properties": {
                        param_name: {
                            "type": type_mapping.get(_type, "string"),  # Default to string if type not in map
                            "description": param_name,  # Or a more descriptive text if available
                        }
                        for param_name, _type in parameters.items()
                    },
                    "required": required_parameters or [],
                    "additionalProperties": False
                }
            }
        }
        self.tools.append(tool)
        self.functions[func.__name__] = func


    def get_messages(self) -> list[Message]:
        return self.messages[-self.message_memory:]

    def last_token_count(self):
        return self.input_token_count, self.output_token_count, self.input_token_count + self.output_token_count

    def prompt(self, prompt, *, images: ImageInput = None,
               return_json=False, response_format=None, cached=True):
        self.reset()
        return self.chat(prompt, images=images, return_json=return_json, response_format=response_format, cached=cached)

    def chat(self, prompt, *, images: ImageInput = None, return_json=False, response_format=None, cached=True):
        if return_json and not self.model.supports_return_json:
            raise NotImplementedError(f"{self.model.model_name} does not support return_json")
        if images and not self.model.supports_image_input:
            raise NotImplementedError(f"{self.model.model_name} does not support image input")

        start_time = time.time()
        if images and not isinstance(images, list):
            images = [images]
        self.append_messages(prompt, images)

        model_response = cached_llm_response(self.model, self.get_messages(), self.tools, return_json=return_json,
                                             response_format=response_format, use_cache=cached)
        result, self.input_token_count, self.output_token_count, tool_use = model_response
        calls = 0  # Safety to prevent infinite loop
        while tool_use and calls < 3:
            # Add the assistant's tool call message to the history
            assistant_message = Message(role='assistant', tool_use=tool_use)
            self.messages.append(assistant_message)

            calls += 1
            function = self.functions.get(tool_use["function_to_call"])
            if not function:
                raise ValueError(f"Function {tool_use['function_to_call']} not found")

            # Run the tool/function
            tool_use['function_result'] = function(*tool_use['function_parameters'].values())

            # Add a model specific message with the tool use result to the conversation
            self.messages.append(self.model.tool_use_message(tool_use))

            model_response = cached_llm_response(
                self.model,
                self.get_messages(),
                self.tools,
                return_json=return_json,
                response_format=response_format,
                use_cache=cached,
            )
            result, self.input_token_count, self.output_token_count, tool_use = model_response

        # Add the result to the messages
        self.messages.append(Message('assistant', str(result)))
        self.last_response_time = time.time() - start_time
        return result
    
    async def prompt_async(self, prompt, *, images: ImageInput = None):
        self.reset()
        # Use 'async for' to properly yield from the chat_async generator
        async for word in self.chat_async(prompt=prompt, images=images):
            yield word

    async def chat_async(self, prompt, *, images: ImageInput = None):
        if images and not isinstance(images, list):
            images = [images]
        self.append_messages(prompt, images)
        for word, _ in self.model.chat_async(messages=self.get_messages()):
            if word:
                yield word

    async def prompt_async_reasoning(self, prompt, *, images: ImageInput = None):
        self.reset()
        # Use 'async for' to properly yield from the chat_async_reasoning generator
        async for word, reasoning_content in self.chat_async_reasoning(prompt=prompt, images=images):
            yield word, reasoning_content

    async def chat_async_reasoning(self, prompt, *, images: ImageInput = None):
        """ Same as chat_async but returns the reasoning content as well
        """
        if images and not isinstance(images, list):
            images = [images]
        self.append_messages(prompt, images)
        for word, reasoning_content in self.model.chat_async(messages=self.get_messages()):
            if word or reasoning_content:
                yield word, reasoning_content

    def after_response(self):
        # content is in messages[-1]['completion']['choices'][0]['message']['content']
        return  # Can be overridden

    def token_count(self, text: str):
        return self.model.token_count(text)



""" Implementation of the Anthropic models. 

Feature table:
    - Async chat:       YES (1)
    - Return JSON:      YES
    - Structured types: NO
    - Token count:      YES
    - Image support:    YES
    - Tool use:         not yet
    
Models:
Claude 3 Opus:	 claude-3-opus-20240229
Claude 3 Sonnet: claude-3-5-sonnet-20240620
Claude 3 Haiku:  claude-3-haiku-20240307

Supported parameters:
max_tokens: int (default 800)
temperature: float (default 0.8)

(1) In contrast to Model.chat, Model.chat_async cannot return json and does not return input and output token counts

"""

import json
import os
from typing import Any

import httpx
from anthropic import Anthropic, AsyncAnthropic, APIConnectionError, AuthenticationError, PermissionDeniedError, \
    APITimeoutError, RateLimitError, BadRequestError
from dotenv import dotenv_values
from google.api_core.exceptions import InternalServerError

from justai.model.message import Message
from justai.models.basemodel import BaseModel, identify_image_format_from_base64, ConnectionException, AuthorizationException, \
    ModelOverloadException, RatelimitException, BadRequestException, GeneralException
from justai.tools.display import ERROR_COLOR, color_print


class AnthropicModel(BaseModel):
    def __init__(self, model_name: str, params: dict):
        system_message = f"You are {model_name}, a large language model trained by Anthropic."
        super().__init__(model_name, params, system_message)
        self.cached_prompt = None

        # Authentication
        if "ANTHROPIC_API_KEY" in params:
            api_key = params["ANTHROPIC_API_KEY"]
            del params["ANTHROPIC_API_KEY"]
        else:
            api_key = os.getenv("ANTHROPIC_API_KEY") or dotenv_values()["ANTHROPIC_API_KEY"]
        if not api_key:
            color_print(
                "No Anthropic API key found. Create one at https://console.anthropic.com/settings/keys and "
                + "set it in the .env file like ANTHROPIC_API_KEY=here_comes_your_key.",
                color=ERROR_COLOR,
            )

        # Client
        timeout = httpx.Timeout(30.0)
        if params.get("async"):
            http_client = httpx.AsyncClient(timeout=timeout)
            self.client = AsyncAnthropic(api_key=api_key, http_client=http_client)
        else:
            http_client = httpx.Client(timeout=timeout)
            self.client = Anthropic(api_key=api_key, http_client=http_client)

        # Required model parameters
        if "max_tokens" not in params:
            params["max_tokens"] = 800

    def chat(self, messages: list[Message], tools, return_json: bool, response_format, log=None) \
            -> tuple[str | Any, int, int, dict[str, object | str] | dict[str, Any]]:
        if response_format:
            raise NotImplementedError("Anthropic does not support response_format")

        antr_messages = transform_messages(messages, return_json)
        antr_tools = transform_tools(tools)
        system_message = self.cached_system_message() if self.cached_prompt else self.system_message

        try:
            message = self.client.messages.create(
                model=self.model_name,
                system=system_message,
                messages=antr_messages,
                tools=antr_tools,
                **self.model_params,
            )
        except APIConnectionError as e:
            print("LLM call failed (APIConnectionError):", repr(e))
            raise ConnectionException(e)
        except (AuthenticationError, PermissionDeniedError) as e:
            print("LLM call failed (Auth):", repr(e))
            raise AuthorizationException(e)
        except InternalServerError as e:
            print("LLM call failed (500):", repr(e))
            raise ModelOverloadException(e)
        except RateLimitError as e:
            print("LLM call failed (RateLimit):", repr(e))
            raise RatelimitException(e)
        except BadRequestError as e:
            print("LLM call failed (BadRequest):", repr(e))
            raise BadRequestException(e)
        except Exception as e:
            print("LLM call failed (Unexpected):", repr(e))
            raise GeneralException(e)

        # Text content
        response_str = message.content[0].text
        if return_json:
            response_str = response_str.split("</json>")[0]  # !!
            try:
                response = json.loads(response_str, strict=False)
            except json.decoder.JSONDecodeError:
                print("ERROR DECODING JSON, RESPONSE:", response_str)
                response = response_str
        else:
            response = response_str

        # Tool use
        # if message.stop_reason == "tool_use":
        #     for c in message.content:
        #         if isinstance(c, ToolUseBlock):
        #             tool_use = {'function_to_call': c.name,
        #                         'function_parameters': c.input,
        #                         'call_id': c.id}
        #             break  # For now, only one function call is supported
        # else:
        #     tool_use = {}
        tool_use = {}

        # Token count
        input_tokens = message.usage.input_tokens
        output_tokens = message.usage.output_tokens
        if self.cached_prompt:
            self.cache_creation_input_tokens = message.usage.cache_creation_input_tokens
            self.cache_read_input_tokens = message.usage.cache_read_input_tokens
        else:
            self.cache_creation_input_tokens = self.cache_read_input_tokens = 0

        return response, input_tokens, output_tokens, tool_use

    def chat_async(self, messages: list[Message]) -> [str, str]:
        try:
            # stream = self.client.messages.create(
            #     model=self.model_name,
            #     max_tokens=self.model_params["max_tokens"],
            #     temperature=self.model_params.get('temperature', 0.8),
            #     system=self.system_message,
            #     messages=transform_messages(messages, return_json=False),
            #     stream=True,
            # )
            stream = self.client.messages.create(
                model=self.model_name,
                system=self.system_message,
                messages=transform_messages(messages, return_json=False),
                stream=True,
                **self.model_params
            )
        except APIConnectionError as e:
            raise ConnectionException(e)
        except (AuthenticationError, PermissionDeniedError) as e:
            raise AuthorizationException(e)
        except APITimeoutError as e:
            raise ModelOverloadException(e)
        except RateLimitError as e:
            raise RatelimitException(e)
        except BadRequestError as e:
            raise BadRequestException(e)
        except Exception as e:
            raise GeneralException(e)

        for event in stream:
            if hasattr(event, "delta") and hasattr(event.delta, "text"):
                yield event.delta.text, None   # 2nd parameter is reasoning_content. Not available yet for Anthropic

    def cached_system_message(self) -> list[dict]:
        return [
                  {
                    "type": "text",
                    "text": self.system_message,
                  },
                  {
                    "type": "text",
                    "text": self.cached_prompt,
                    "cache_control": {"type": "ephemeral"}
                  }
                ]

    @staticmethod  # !! Veranderen als deze niet anders wordt dan die van OpenAI
    def tool_use_message(tool_use) -> Message:
        return Message(role='user', content='', tool_use=tool_use)

    def token_count(self, text: str) -> int:
        messages = transform_messages([Message("user", text)], return_json=False)
        return self.client.beta.messages.count_tokens(model=self.model_name, messages=messages)


def transform_messages(messages: list[Message], return_json: bool) -> list[dict]:
    # Anthropic requires the first message to be a user message
    msgs = messages[next(i for i, message in enumerate(messages) if message.role == "user"):]

    if msgs and return_json:
        msgs += [Message("assistant", "<json>")]
    result = [create_anthropic_message(msg) for msg in msgs]
    return result


def transform_tools(tools: list[dict]) -> list[dict]:
    """
    At Anthropic tools work like this:
    tools=[
        {
            "name": "get_weather",
            "description": "Get the current weather in a given location",
            "input_schema": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    }
                },
                "required": ["location"],
            },
        }
    ],
    """
    def transform_tool(tool:dict) -> dict:
        if "function" in tool:
            tool["name"] = tool["function"]["name"]
            tool["description"] = tool["function"]["description"]
            tool["input_schema"] = tool["function"]["parameters"]
            del tool["function"]
        if "type" in tool:
            del tool["type"]
        if "function" in tool:
            del tool["function"]
        return tool

    return [transform_tool(tool) for tool in tools]


def create_anthropic_message(message: Message):

    content = []
    role = message.role

    for img in message.images:
        base64img = Message.to_base64_image(img)
        mime_type = identify_image_format_from_base64(base64img)
        content += [
            {
                "type": "image",
                "source": {
                    "type": "base64",
                    "media_type": mime_type,
                    "data": base64img,
                },
            }
        ]

    if message.content:
        content += [{"type": "text", "text": message.content}]

    if message.tool_use:
        _input = message.tool_use["function_result"]
        if not isinstance(input, dict): # Anthropic requires input to be a dict
            _input = {'data': _input}
        content += [{
            "type": "tool_use",
            "id": message.tool_use["call_id"],
            "name": message.tool_use["function_to_call"],
            "input": _input,
        }]
        role = 'assistant' # Anthropic requires the role to be assistant when using tools

    return {"role": role, "content": content}



import os

from dotenv import dotenv_values
from openai import OpenAI

from justai.model.message import Message
from justai.models.basemodel import BaseModel
from justai.models.openai_models import OpenAIModel
from justai.tools.display import color_print, ERROR_COLOR


class PerplexityModel(OpenAIModel):
    def __init__(self, model_name: str, params: dict = None):
        system_message = f"You are {model_name}, a large language model trained by Perplexity."
        BaseModel.__init__(self, model_name, params, system_message)

        # Authentication
        keyname = "PERPLEXITY_API_KEY"
        api_key = params.get(keyname) or os.getenv(keyname) or dotenv_values()[keyname]
        if not api_key:
            color_print(f"No {keyname} found. Create one at https://www.perplexity.ai/settings/api and " +
                        f"set it in the .env file like {keyname}=here_comes_your_key.", color=ERROR_COLOR)
        self.client = OpenAI(api_key=api_key, base_url="https://api.perplexity.ai")

        # Overwrite parent class defaults
        self.supports_return_json = False

    def chat_async(self, messages: list[Message]):
        """ Perplexity does not separately return thinking content
        but returns it's thinking between <think> and </think> tags."""
        thinking = False
        for content, _ in super().chat_async(messages):
            if content == '<think>':
                thinking = True
            elif '</think>' in content:
                thinking = False
            elif thinking:
                yield None, content
            else:
                yield content, None



import os

from dotenv import dotenv_values
from openai import OpenAI

from justai.models.basemodel import BaseModel
from justai.models.openai_models import OpenAIModel
from justai.tools.display import color_print, ERROR_COLOR


class OpenRouterModel(OpenAIModel):
    def __init__(self, model_name: str, params: dict = None):
        system_message = f"You are {model_name}, a large language model."
        BaseModel.__init__(self, model_name, params, system_message)

        # Authentication
        keyname = "OPENROUTER_API_KEY"
        api_key = params.get(keyname) or os.getenv(keyname) or dotenv_values()[keyname]
        if not api_key:
            color_print("No X AI API key found. Create one at https://openrouter.ai/settings/keys and " +
                        f"set it in the .env file like {keyname}=here_comes_your_key.", color=ERROR_COLOR)
        self.client = OpenAI(api_key=api_key, base_url="https://openrouter.ai/api/v1")
